"""
List of objects

.. code-block:: python

    class Doc(Document):
        __collection__ = 'docs'
        integers = fields.ListField(fields.IntegerField)

    doc = Doc()
    doc.integers.append(1)
    doc.integers.append(2)
    assert doc.integers == [1, 2]

    db.insert(doc)
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload

    doc.integers.append(3)  # do not save
    assert doc.integers == [1, 2, 3]
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.integers == [1, 2]

    doc.integers.remove(2)  # do not save too
    assert doc.integers == [1]
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.integers == [1, 2]

    doc.integers.push(3)  # $push query
    assert doc.integers == [1, 2, 3]
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.integers == [1, 2, 3]

    doc.integers.pull(2)  # $pull query
    assert doc.integers == [1, 3]
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.integers == [1, 3]

"""
from collections import abc

from yadm.fields.containers import (
    ArrayContainer,
    ArrayField,
)


class List(ArrayContainer, abc.MutableSequence):
    """ Container for list
    """
    def __repr__(self):
        return 'List({!r})'.format(self._data)

    def _load_from_mongo(self, data):
        self._data = []

        for item in data or ():
            if hasattr(self._field.item_field, 'from_mongo'):
                value = self._field.item_field.from_mongo(self.__parent__, item)
            else:
                value = self._prepare_value(item)

            self._data.append(value)

    def insert(self, index, item):
        """ Append item to list

        :param item: item for append

        This method does not save object!
        """
        self._data.insert(index, self._prepare_value(item))
        self._set_changed()

    def append(self, item):
        """ Append item to list

        :param item: item for append

        This method does not save object!
        """
        self._data.append(self._prepare_value(item))
        self._set_changed()

    def remove(self, item):
        """ Remove item from list

        :param item: item for remove

        This method does not save object!
        """
        self._data.remove(item)
        self._set_changed()

    def push(self, item):
        """ Push item directly to database

        :param item: item for `$push`

        See `$push` in MongoDB's `update`.
        """
        item = self._prepare_value(item)

        if hasattr(self._field.item_field, 'to_mongo'):
            data = self._field.item_field.to_mongo(self.__document__, item)
        else:
            data = item

        qs = self._get_queryset()
        qs.update({'$push': {self.__field_name__: data}}, multi=False)
        self._data.append(item)

    def pull(self, query, reload=True):
        """ Pull item from database

        :param query: query for `$pull` on this field
        :param bool reload: automatically reload all values from database

        See `$pull` in MongoDB's `update`.
        """
        qs = self._get_queryset()
        qs.update({'$pull': {self.__field_name__: query}}, multi=False)

        if reload:
            doc = qs.find_one()
            self._load_from_mongo(self.__get_value__(doc))

    def replace(self, query, item, reload=True):
        """ Replace list elements

        :param query: query for `update`.
            Keys of this query is relative.
        :param item: embedded document or dict
        """
        if hasattr(self._field.item_field, 'to_mongo'):
            data = self._field.item_field.to_mongo(self.__document__, item)
        else:
            data = item

        processed_query = {}
        for key, value in query.items():
            processed_query['.'.join([self.__field_name__, key])] = value

        qs = self._get_queryset()
        qs = qs.find(processed_query)
        qs.update({'$set': {'.'.join([self.__field_name__, '$']): data}})

        if reload:
            doc = self._get_queryset().find_one()
            self._load_from_mongo(self.__get_value__(doc))

    def update(self, query, values, reload=True):
        """ Update fields in embedded documents

        :param query: query for `update`.
            Keys of this query is relative.
        :param values: dict of new values
        """
        processed_query = {}
        for key, value in query.items():
            processed_query['.'.join([self.__field_name__, key])] = value

        data = {}
        for key, value in values.items():
            data['.'.join([self.__field_name__, '$', key])] = value

        qs = self._get_queryset()
        qs = qs.find(processed_query)
        qs.update({'$set': data})

        if reload:
            doc = self._get_queryset().find_one()
            self._load_from_mongo(self.__get_value__(doc))


class ListField(ArrayField):
    """ Field for list values

    For example, document with list of integers:

    .. code-block:: python

        class TestDoc(Document):
            __collection__ = 'testdoc'
            li = fields.ListField(fields.IntegerField())
    """
    container = List

    @property
    def default(self):
        return []



from yadm import fields
from yadm.documents import Document, EmbeddedDocument

from .test_database import BaseDatabaseTest


class SetEmbeddedFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestEDoc(EmbeddedDocument):
            i = fields.IntegerField()

        class TestDoc(Document):
            __collection__ = 'testdocs'
            li = fields.SetField(fields.EmbeddedDocumentField(TestEDoc))

        self.TestEDoc = TestEDoc
        self.TestDoc = TestDoc

    def test_save(self):
        td = self.TestDoc()

        ted = self.TestEDoc()
        ted.i = 13
        td.li.add(ted)

        ted = self.TestEDoc()
        ted.i = 42
        td.li.add(ted)

        self.db.insert(td)

        data = self.db.db.testdocs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 2)
        self.assertIn('i', data['li'][0])

        res = set()

        for item in data['li']:
            self.assertIn('i', item)
            self.assertIsInstance(item['i'], int)
            res.add(item['i'])

        self.assertEquals(res, {13, 42})

    def test_load(self):
        _id = self.db.db.testdocs.insert({'li': []})
        self.db.db.testdocs.update({'_id': _id}, {'$addToSet': {'li': {'i': 13}}})
        self.db.db.testdocs.update({'_id': _id}, {'$addToSet': {'li': {'i': 42}}})

        doc = self.db.get_queryset(self.TestDoc).find_one()

        self.assertTrue(hasattr(doc, 'li'))
        self.assertEqual(len(doc.li), 2)

        res = set()

        for item in doc.li:
            self.assertIsInstance(item, self.TestEDoc)
            self.assertTrue(hasattr(item, 'i'))
            self.assertIsInstance(item.i, int)
            res.add(item.i)

        self.assertEquals(res, {13, 42})

    def test_add_to_set(self):
        td = self.TestDoc()
        self.db.insert(td)

        ted = self.TestEDoc()
        ted.i = 13
        td.li.add_to_set(ted)

        ted = self.TestEDoc()
        ted.i = 42
        td.li.add_to_set(ted)

        data = self.db.db.testdocs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 2)
        self.assertIn('i', data['li'][0])
        self.assertEqual(data['li'][0]['i'], 13)
        self.assertEqual(data['li'][1]['i'], 42)



"""
Basic documents classes for build models.

.. code-block:: python

    class User(Document):
        __collection__ = 'users'

        first_name = fields.StringField
        last_name = fields.StringField
        age = fields.IntegerField


All fields placed in :py:mod:`yadm.fields` package.
"""

from bson import ObjectId

from yadm.markers import AttributeNotSet, NoDefault
from yadm.fields.base import Field
from yadm.fields.simple import ObjectIdField


class MetaDocument(type):
    '''Metaclass for documents'''
    def __init__(cls, name, bases, cls_dict):
        cls.__fields__ = {}

        for base in bases:
            if isinstance(base, MetaDocument):
                for name, field in base.__fields__.items():
                    if name not in cls_dict:
                        cls_dict[name] = field.copy()

        for attr, field in cls_dict.items():
            if isinstance(field, type) and issubclass(field, Field):
                field = field()

            if hasattr(field, 'contribute_to_class'):
                field.contribute_to_class(cls, attr)

            else:
                setattr(cls, attr, field)

        super().__init__(name, bases, cls_dict)


class BaseDocument(metaclass=MetaDocument):
    """ Base class for all documents
    """
    __initialized__ = False

    def __init__(self, *args, **kwargs):
        if args:
            if len(args) != 1:
                raise TypeError("only one positional argument accepted!")

            elif not isinstance(args[0], dict):
                name = type(args[0]).__name__
                raise TypeError("argument must be a dict, not {}".format(name))

            data = args[0]

        elif kwargs:
            data = kwargs

        else:
            data = {}

        self.__data__ = {}
        self.__fields_changed__ = set()

        for key, field in self.__fields__.items():
            if key in data:
                setattr(self, key, data[key])
            else:
                default = field.default

                if default is NoDefault:
                    self.__data__[key] = AttributeNotSet
                else:
                    setattr(self, key, default)

        self.__initialized__ = data.get('__initialized__', True)

    def __str__(self):
        """ Implement it for pretty str and repr documents
        """
        return str(id(self))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, str(self))


class Document(BaseDocument):
    """ Class for build first level documents

    .. py:attribute:: __collection__

        Name of MongoDB collection

    .. py:attribute:: _id

        Mongo object id (:py:class:`bson.ObjectId`)

    .. py:attribute:: id

        Alias for :py:attr:`_id` for simply use

    .. py:attribute:: __db__

        Internal attribute contain instance of :py:class:`yadm.database.Database`
        for realize :py:class:`yadm.fields.references.ReferenceField`.
        It bind in :py:class:`yadm.database.Database` or :py:class:`yadm.queryset.QuerySet`.
    """
    __collection__ = None
    __db__ = None

    _id = ObjectIdField()

    def __str__(self):
        if hasattr(self, '_id'):
            return '{!s}:{!s}'.format(self.__collection__, self._id)
        else:
            return '{!s}:<empty>'.format(self.__collection__)

    def __eq__(self, other):
        if isinstance(other, Document):
            return self.id == other.id
        elif isinstance(other, ObjectId):
            return self.id == other
        else:
            return False

    def __hash__(self):
        return hash(self.id)

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, id):
        self._id = id

    @id.deleter
    def id(self, id):
        del self._id


class DocumentItemMixin:
    """ Mixin for custom all fields values,
    such as :py:class:`EmbeddedDocument`, :py:class:`yadm.fields.containers.Container`

    .. py:attribute:: __parent__

        Parent object.

        .. code-block:: python

            assert doc.embedded_doc.__parent__ is doc
            assert doc.list[13].__parent__ is doc.list

    .. py:attribute:: __name__

        .. code-block:: python

            assert doc.list.__name__ == 'list'
            assert doc.list[13].__name__ == 13

    """
    __parent__ = None
    __name__ = None

    @property
    def __document__(self):
        """ Root document

        .. code-block:: python

                assert doc.f.l[0].__document__ is doc
        """
        obj = self

        while getattr(obj, '__parent__', None):
            obj = obj.__parent__

        return obj

    @property
    def __db__(self):
        """ Database object

        .. code-block:: python

            assert doc.f.l[0].__db__ is doc.__db__
        """
        return self.__document__.__db__

    @property
    def __path__(self):
        """ Path to root generator

        .. code-block:: python

            assert list(doc.f.l[0].__path__) == [doc.f.l[0], doc.f.l, doc.f]
        """
        obj = self

        while getattr(obj, '__parent__', None):
            yield obj
            obj = obj.__parent__

    @property
    def __path_names__(self):
        """ Path to root generator

        .. code-block:: python

            assert list(doc.f.l[0].__path__) == [0, 'l', 'f']
        """
        for item in self.__path__:
            yield item.__name__

    @property
    def __field_name__(self):
        """ Dotted field name for MongoDB opperations,
        like as $set, $push and other...

        .. code-block:: python

            assert doc.f.l[0].__field_name__ == 'f.l.0'
        """
        return '.'.join(reversed([str(i) for i in self.__path_names__]))

    def __get_value__(self, document):
        """ Get value from document with path to self
        """
        obj = document

        for name in reversed(list(self.__path_names__)):
            if isinstance(name, int):
                obj = obj[name]
            else:
                obj = getattr(obj, name)

        return obj


class EmbeddedDocument(DocumentItemMixin, BaseDocument):
    """ Class for build embedded documents
    """
    def __init__(self, *args, **kwargs):
        if '__parent__' in kwargs:
            self.__parent__ = kwargs.pop('__parent__')

        if '__name__' in kwargs:
            self.__name__ = kwargs.pop('__name__')

        super().__init__(*args, **kwargs)



"""
Field with sets.

Similar as :py:mod:`yadm.fields.list`.
"""
from collections import abc

from yadm.fields.containers import (
    ArrayContainer,
    ArrayField,
)


class Set(ArrayContainer, abc.MutableSet):
    """ Container for set
    """
    def __repr__(self):
        return 'Set({!r})'.format(self._data)

    def _load_from_mongo(self, data):
        self._data = set()

        for item in data or ():
            if hasattr(self._field.item_field, 'from_mongo'):
                value = self._field.item_field.from_mongo(self.__document__, item)
            else:
                value = self._prepare_value(item)

            self._data.add(value)

    def add(self, item):
        """ Append item to set

        :param item: item for add

        This method does not save object!
        """
        self._data.add(self._prepare_value(item))
        self._set_changed()

    def discard(self, item):
        """ Remove item from the set if it is present

        :param item: item for discard

        This method does not save object!
        """
        self._data.discard(item)
        self._set_changed()

    def remove(self, item):
        """ Remove item from set

        :param item: item for remove

        This method does not save object!
        """
        self._data.remove(item)
        self._set_changed()

    def add_to_set(self, item):
        """ Add item directly to database

        :param item: item for `$addToSet`

        See `$addToSet` in MongoDB's `update`.
        """
        item = self._prepare_value(item)

        if hasattr(self._field.item_field, 'to_mongo'):
            data = self._field.item_field.to_mongo(self.__document__, item)
        else:
            data = item

        qs = self._get_queryset()
        qs.update({'$addToSet': {self.__field_name__: data}}, multi=False)
        self._data.add(item)

    def pull(self, query, reload=True):
        """ Pull item from database

        :param query: query for `$pull` on this field
        :param bool reload: automatically reload all values from database

        See `$pull` in MongoDB's `update`.
        """
        qs = self._get_queryset()
        qs.update({'$pull': {self.__field_name__: query}}, multi=False)

        if reload:
            doc = qs.find_one()
            self._load_from_mongo(self.__get_value__(doc))


class SetField(ArrayField):
    """ Field for set values
    """
    container = Set

    @property
    def default(self):
        return set()



from unittest import TestCase

from yadm import fields
from yadm.documents import Document

from .test_database import BaseDatabaseTest


class MoneyTest(TestCase):
    def test_str(self):
        self.assertEqual(str(fields.Money('3.14')), '3.14')
        self.assertEqual(str(fields.Money('0.14')), '0.14')
        self.assertEqual(str(fields.Money('30.14')), '30.14')
        self.assertEqual(str(fields.Money('3.1')), '3.10')
        self.assertEqual(str(fields.Money('0.1')), '0.10')

    def test_to_mongo(self):
        self.assertEqual(fields.Money('3.14').to_mongo(), 314)
        self.assertEqual(fields.Money('0.14').to_mongo(), 14)
        self.assertEqual(fields.Money('30.14').to_mongo(), 3014)
        self.assertEqual(fields.Money('3.1').to_mongo(), 310)
        self.assertEqual(fields.Money('0.1').to_mongo(), 10)


class MoneyFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdocs'
            money = fields.MoneyField()

        self.TestDoc = TestDoc

    def test_save(self):
        doc = self.TestDoc()
        doc.money = fields.Money('3.14')
        self.db.save(doc)

        data = self.db.db.testdocs.find_one()

        self.assertIn('money', data)
        self.assertIsInstance(data['money'], int)
        self.assertEqual(data['money'], 314)

    def test_load(self):
        self.db.db.testdocs.insert({'money': 314})

        doc = self.db.get_queryset(self.TestDoc).find_one()

        self.assertTrue(hasattr(doc, 'money'))
        self.assertIsInstance(doc.money, fields.Money)
        self.assertEqual(doc.money, fields.Money('3.14'))

    def test_load_and_save(self):
        self.db.db.testdocs.insert({'money': 314})

        doc = self.db.get_queryset(self.TestDoc).find_one()
        self.db.save(doc)

        self.assertTrue(hasattr(doc, 'money'))
        self.assertIsInstance(doc.money, fields.Money)
        self.assertEqual(doc.money, fields.Money('3.14'))

        data = self.db.db.testdocs.find_one()

        self.assertIn('money', data)
        self.assertIsInstance(data['money'], int)
        self.assertEqual(data['money'], 314)



from yadm.documents import Document
from yadm import fields

from .test_database import BaseDatabaseTest


class JoinTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        self.id_ref_1 = self.db.db.testdocs_ref.insert({'i': 1})
        self.id_ref_2 = self.db.db.testdocs_ref.insert({'i': 2})

        for n in range(10):
            self.db.db.testdocs.insert({
                'i': n,
                'ref': self.id_ref_1 if n % 2 else self.id_ref_2,
            })

        class TestDocRef(Document):
            __collection__ = 'testdocs_ref'
            i = fields.IntegerField

        class TestDoc(Document):
            __collection__ = 'testdocs'
            i = fields.IntegerField
            ref = fields.ReferenceField(TestDocRef)

        self.TestDocRef = TestDocRef
        self.TestDoc = TestDoc

        self.qs = self.db(self.TestDoc)
        self.join = self.qs.join('ref')

    def test_len(self):
        self.assertEqual(len(self.join), 10)

    def test_iter(self):
        self.assertEqual(len(list(self.join)), 10)

    def test_join(self):
        for doc in self.join:
            self.assertIn('ref', doc.__data__)
            ref = doc.__data__['ref']
            self.assertIsInstance(ref, Document)
            self.assertEqual(doc.ref.id, self.id_ref_1 if doc.i % 2 else self.id_ref_2)

    def test_get_queryset(self):
        self.db.db.testdocs_ref.insert({'i': 3})
        qs = self.qs.join().get_queryset('ref')
        self.assertEqual(qs.count(), 2)
        self.assertEqual(
            {d.id for d in qs},
            {self.id_ref_1, self.id_ref_2},
        )


class ManyJoinTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        self.id_ref_1_1 = self.db.db.testdocs_ref_1.insert({'i': 11})
        self.id_ref_1_2 = self.db.db.testdocs_ref_1.insert({'i': 12})

        self.id_ref_2_1 = self.db.db.testdocs_ref_2.insert({'i': 21})
        self.id_ref_2_2 = self.db.db.testdocs_ref_2.insert({'i': 22})

        for n in range(10):
            self.db.db.testdocs.insert({
                'i': n,
                'ref_1': self.id_ref_1_1 if n % 2 else self.id_ref_1_2,
                'ref_2': self.id_ref_2_1 if n % 2 else self.id_ref_2_2,
            })

        class TestDocRef(Document):
            i = fields.IntegerField

        class TestDocRef_1(Document):
            __collection__ = 'testdocs_ref_1'

        class TestDocRef_2(Document):
            __collection__ = 'testdocs_ref_2'

        class TestDoc(Document):
            __collection__ = 'testdocs'
            i = fields.IntegerField
            ref_1 = fields.ReferenceField(TestDocRef_1)
            ref_2 = fields.ReferenceField(TestDocRef_2)

        self.TestDoc = TestDoc

        self.qs = self.db(self.TestDoc)
        self.join = self.qs.join('ref_1', 'ref_2')

    def test_join(self):
        for doc in self.join:
            for n in range(1, 3):
                fn = 'ref_{}'.format(n)
                self.assertIn(fn, doc.__data__)
                ref = doc.__data__[fn]
                self.assertIsInstance(ref, Document)
                self.assertEqual(
                    getattr(doc, fn).id,
                    getattr(self, ('id_{}_1'.format(fn) if doc.i % 2 else 'id_{}_2'.format(fn)))
                )



from yadm import fields
from yadm.documents import Document
from yadm.markers import NotLoaded

from .test_database import BaseDatabaseTest


class DeferredLoadTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdocs'

            i = fields.IntegerField()
            s = fields.StringField()

        self.TestDoc = TestDoc

    def test(self):
        _id = self.db.db.testdocs.insert({'i': 13, 's': 'test'})
        doc = self.db.get_queryset(self.TestDoc).fields('i').with_id(_id)

        self.assertEqual(doc.__data__, {'_id': _id, 'i': 13, 's': NotLoaded})

        self.assertTrue(hasattr(doc, 'i'))
        self.assertEqual(doc.i, 13)

        self.assertTrue(hasattr(doc, 's'))
        self.assertEqual(doc.s, 'test')



from collections import abc
from collections import defaultdict

from bson import ObjectId

from yadm.documents import Document
from yadm.fields.reference import ReferenceField


class Join(abc.Sequence):
    """ Helper for build joins
    """
    def __init__(self, qs):
        self._qs = qs
        self._document_class = qs._document_class
        self._db = qs._db

        self._indexes = defaultdict(dict)
        self._map_name_type = {}
        self._map_type_names = defaultdict(set)
        self._map_name_ids = defaultdict(set)

        self._data = list(qs)

    # abc.Sequence method

    def __iter__(self):
        return (i for i in self._data)

    def __getitem__(self, idx):
        return self._data[idx]

    def __contains__(self, item):
        return item in self._data

    def __len__(self):
        return len(self._data)

    def __reversed__(self):
        return reversed(self._data)

    def index(self, item):
        return self._data.index(item)

    # end abc.Sequence

    def get_queryset(self, field_name):
        """ Return queryset for joined objects
        """
        field = self._get_field(field_name)
        qs = self._db(field.reference_document_class)
        ids = self._get_joined_ids(field_name)
        return qs.find({'_id': {'$in': list(ids)}})

    def join(self, *field_names):
        """ Do manual join
        """
        self._load_names_types_maps(*field_names)
        self._load_map_name_ids(*field_names)
        self._load_objects_to_indexes(*field_names)
        self._set_objects(*field_names)

    def _get_field(self, field_name):
        document_fields = self._document_class.__fields__
        if field_name not in document_fields:
            raise ValueError(field_name)

        field = document_fields[field_name]

        if not isinstance(field, ReferenceField):
            raise RuntimeError('bad field type: {!r}'.format(field_name))

        return field

    def _get_joined_ids(self, field_name):
        ids = set()
        for doc in self:
            value = doc.__data__[field_name]

            _id = self._prepare_id(value)
            if _id:
                ids.add(_id)

        return ids

    def _load_names_types_maps(self, *field_names):
        for field_name in field_names:
            field = self._get_field(field_name)
            reference_document_class = field.reference_document_class
            self._map_name_type[field_name] = reference_document_class
            self._map_type_names[reference_document_class].add(field_name)

    def _load_map_name_ids(self, *field_names):
        for doc in self:
            for field_name in field_names:
                value = doc.__data__.get(field_name)

                _id = self._prepare_id(value)
                if _id:
                    self._map_name_ids[field_name].add(_id)

    def _load_objects_to_indexes(self, *field_names):
        field_names = set(field_names)

        for joined_document_class, names in self._map_type_names.items():
            if not (field_names & names):
                continue

            ids = set()
            for field_name in names:
                ids.update(self._map_name_ids[field_name])

            ids = list(ids - set(self._indexes[joined_document_class]))
            qs = self._db(joined_document_class).find({'_id': {'$in': ids}})
            self._indexes[joined_document_class].update(qs.bulk())

    def _set_objects(self, *field_names):
        for doc in self:
            for field_name in field_names:
                value = doc.__data__.get(field_name)

                _id = self._prepare_id(value)
                if _id:
                    joined_document_class = self._map_name_type[field_name]
                    index = self._indexes[joined_document_class]
                    doc.__data__[field_name] = index.get(_id, value)

    def _prepare_id(self, value):
        if isinstance(value, ObjectId):
            return value
        elif isinstance(value, Document):
            return value.id



from unittest import TestCase

from yadm.documents import Document
from yadm.fields.email import EmailField


class EmailFieldTest(TestCase):
    def test_ok(self):
        class TestDoc(Document):
            e = EmailField()

        td = TestDoc()
        td.e = 'E@mA.iL'

        self.assertEqual(td.e, 'e@ma.il')

    def test_error(self):
        class TestDoc(Document):
            e = EmailField()

        td = TestDoc()

        with self.assertRaises(ValueError):
            td.e = 'EmA.iL'



"""
Base classes for containers.
"""

from yadm.fields import FieldDescriptor, Field
from yadm.documents import DocumentItemMixin


class ContainerDescriptor(FieldDescriptor):
    """ Descriptor for containers
    """
    def __get__(self, instance, owner):
        value = super().__get__(instance, owner)

        if (instance is not None
                and not isinstance(instance.__data__[self.name], self.field.__class__)):

            instance.__data__[self.name] = value

        return value


class Container(DocumentItemMixin):
    """ Base class for containers
    """
    def __init__(self, parent, field, data):
        self.__parent__ = parent
        self._field = field
        self._field_name = field.name
        self._load_from_mongo(data)

    def __iter__(self):
        for n, item in enumerate(self._data):
            if isinstance(item, DocumentItemMixin):
                item.__parent__ = self
                item.__name__ = n

            yield item

    def __getitem__(self, item):
        value = self._data[item]

        if isinstance(value, DocumentItemMixin):
            value.__parent__ = self
            value.__name__ = item

        return value

    def __setitem__(self, item, value):
        self._data[item] = self._prepare_value(value)
        self._set_changed()

    def __delitem__(self, item):
        del self._data[item]
        self._set_changed()

    def __contains__(self, item):
        return item in self._data

    def __len__(self):
        return len(self._data)

    def __bool__(self):
        return bool(self._data)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._data == other._data
        else:
            return self._data == other

    def _prepare_value(self, item):
        """ `prepare_value` function for `item_field`
        """
        raise NotImplementedError

    def _load_from_mongo(self, data):
        """ Load data from pymongo

        Simply, here create self._data with deserialized data from mongo.
        """
        self._data = data

    def _get_queryset(self):
        """ Return queryset for got data for this field
        """
        if self.__db__ is None:
            raise RuntimeError('object not binded to database')

        qs = self.__db__.get_queryset(self.__document__.__class__)
        qs = qs.find({'_id': self.__document__.id})
        return qs.fields(self.__field_name__)

    def _set_changed(self):
        """ Add field to __fields_changed__

        TODO: work with deep objects
        """
        self.__parent__.__fields_changed__.add(self._field_name)


class ArrayContainer(Container):
    def __getitem__(self, item):
        if isinstance(item, slice):
            res = []
            start = item.start
            step = item.step

            for n, value in enumerate(self._data[item]):
                if isinstance(value, DocumentItemMixin):
                    value.__parent__ = self
                    value.__name__ = start + (step * n)

                res.append(value)

            return res

        else:
            return super().__getitem__(item)

    def _prepare_value(self, item):
        """ `prepare_value` function for `item_field`
        """
        return self._field.item_field.prepare_value(None, item)


class ContainerField(Field):
    """ Base class for container fields
    """
    descriptor_class = ContainerDescriptor

    def to_mongo(self, document, value):
        """ Serialize field value to data for MongoDB

        Must be implemented in field class.

        :param BaseDocument document: document with this value
        :param value: python value
        :return: MongoDB data
        """
        return value

    def from_mongo(self, document, value):
        """ Deserialize field value from MongoDB data

        Must be implemented in field class.

        :param BaseDocument document: document with this value
        :param value: MongoDB data
        :return: python value
        """
        raise NotImplementedError


class ArrayField(ContainerField):
    """ Base class for array containers like as lists or set
    """
    container = ArrayContainer

    def __init__(self, item_field):
        if isinstance(item_field, type):
            item_field = item_field()

        self.item_field = item_field

    def to_mongo(self, document, value):
        """ See :py:meth:`ContainerField.to_mongo`
        """
        result = []

        for item in value:
            if hasattr(self.item_field, 'to_mongo'):
                result.append(self.item_field.to_mongo(document, item))
            else:
                result.append(item)

        return result

    def from_mongo(self, document, value):
        """ See :py:meth:`ContainerField.from_mongo`
        """
        return self.container(document, self, value)



"""
Field for decimal numbers

.. code block: python

    class DocClass(Document):
        dec = DecimalField()

    doc = DocClass()
    doc.dec = Decimal('3.14')

    db.insert(doc)

This code save to MongoDB document:

.. code block: javascript

    {
        id: ObjectId('534272984c78591787e1a964'),
        dec: {i: 314, e: -2}
    }

"""
from decimal import Decimal, getcontext
from functools import reduce

from yadm.fields.base import Field, DefaultMixin
from yadm.markers import NoDefault


class DecimalField(DefaultMixin, Field):
    """ Field for work with :class:`decimal.Decimal`

    :param decimal.Context context: context for decimal operations
        (default: run :func:`decimal.getcontext` when need)
    :param decimal.Decimal default:
    """
    def __init__(self, context=None, default=NoDefault):
        self.context = context
        super().__init__(default=default)

    @property
    def context(self):
        """ Context

        :return: :class:`decimal.Context` for values
        """
        return self._context if self._context else getcontext()

    @context.setter
    def context(self, value):
        self._context = value

    @staticmethod
    def _integer_from_digits(digits):
        """ Make integer from digits

        :param list digits: list of digits as integers
        :return: result integer

        [2, 4, 6, 7] => 2467
        """
        return reduce(lambda cur, acc: cur * 10 + acc, digits, 0)

    def prepare_value(self, document, value):
        """ Cast value to :class:`decimal.Decimal`
        """
        if value is None:
            return None
        elif isinstance(value, Decimal):
            return value
        elif isinstance(value, (str, int, float)):
            return Decimal(value, context=self.context)
        else:
            sign = value['i'] < 0  # False - positive, True - negative

            digits = []
            i = abs(value['i'])

            while i:
                i, d = divmod(i, 10)
                digits.append(d)
            digits.reverse()

            return Decimal((sign, digits, value['e']), context=self.context)

    def to_mongo(self, document, value):
        if value is None:
            return None
        else:
            sign, digits, exp = value.as_tuple()
            integer = self._integer_from_digits(digits)
            return {
                'i': -integer if sign else integer,
                'e': exp
            }

    def from_mongo(self, document, data):
        return self.prepare_value(document, data)



from yadm import fields
from yadm.documents import Document

from .test_database import BaseDatabaseTest


class ListFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdoc'
            li = fields.ListField(fields.IntegerField())

        self.TestDoc = TestDoc

    def test_default(self):
        td = self.TestDoc()
        self.assertIsInstance(td.li, fields.list.List)
        self.assertFalse(td.li)
        self.assertEqual(len(td.li), 0)
        self.assertEqual(td.li._data, [])
        self.assertEqual(td.li, [])

    def test_get(self):
        _id = self.db.db.testdoc.insert({'li': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)

        self.assertTrue(td.li)
        self.assertEqual(len(td.li), 3)
        self.assertEqual(td.li._data, [1, 2, 3])
        self.assertEqual(list(td.li), [1, 2, 3])
        self.assertEqual(td.li[1], 2)

    def test_append(self):
        _id = self.db.db.testdoc.insert({'li': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.li.append(4)

        self.assertEqual(td.li, [1, 2, 3, 4])

    def test_append_valueerror(self):
        td = self.TestDoc()
        self.assertRaises(ValueError, td.li.append, 'not a number')

    def test_append_save(self):
        _id = self.db.db.testdoc.insert({'li': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.li.append(4)
        self.db.save(td)

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['li'], [1, 2, 3, 4])

    def test_remove(self):
        _id = self.db.db.testdoc.insert({'li': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.li.remove(2)

        self.assertEqual(td.li, [1, 3])

    def test_remove_save(self):
        _id = self.db.db.testdoc.insert({'li': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.li.remove(2)
        self.db.save(td)

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['li'], [1, 3])

    def test_push(self):
        _id = self.db.db.testdoc.insert({'li': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.li.push(4)

        self.assertEqual(td.li, [1, 2, 3, 4])

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['li'], [1, 2, 3, 4])

    def test_push_valueerror(self):
        td = self.TestDoc()
        self.assertRaises(ValueError, td.li.push, 'not a number')

    def test_pull(self):
        _id = self.db.db.testdoc.insert({'li': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.li.pull(2)

        self.assertEqual(td.li, [1, 3])

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['li'], [1, 3])



"""
Map

.. code-block:: python

    class Doc(Document):
        __collection__ = 'docs'
        map = fields.MapField(fields.IntegerField)

    doc = Doc()
    doc.map['a'] = 1
    doc.map['b'] = 2
    assert doc.map == {'a': 1, 'b': 2}

    db.insert(doc)
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload

    doc.map['c'] = 3  # do not save
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

    del doc.map['b']  # do not save too
    assert doc.map == {'a': 1}
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

    doc.map.set('d', 3)  # $set query
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}

    doc.map.unset('d', 3)  # $unset query
    assert doc.map == {'a': 1, 'b': 2}
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

"""
from collections import abc

from yadm.fields.containers import (
    Container,
    ContainerField,
)


class Map(Container, abc.MutableMapping):
    """ Map
    """

    def __repr__(self):
        return 'Map({!r})'.format(self._data)

    def _load_from_mongo(self, data):
        self._data = {}

        for key, value in (data or {}).items():
            if hasattr(self._field.value_field, 'from_mongo'):
                value = self._field.value_field.from_mongo(self.__parent__, value)
            else:
                value = self._prepare_value(value)

            self._data[key] = value

    def _prepare_value(self, value):
        """ `prepare_value` function for `value_field`
        """
        if hasattr(self._field.value_field, 'prepare_value'):
            return self._field.value_field.prepare_value(None, value)
        else:
            return value

    def set(self, key, value):
        """ Set key directly in database

        :param key: key
        :param value: value for `$set`

        See `$set` in MongoDB's `set`.
        """
        key = str(key)

        if hasattr(self._field.value_field, 'to_mongo'):
            value = self._field.value_field.to_mongo(self.__document__, value)
        else:
            value = self._prepare_value(value)

        qs = self._get_queryset()
        fn = '.'.join([self.__field_name__, key])
        qs.update({'$set': {fn: value}}, multi=False)
        self._data[key] = value

    def unset(self, key):
        """ Unset key directly in database

        :param key: key

        See `$unset` in MongoDB's `unset`.
        """
        key = str(key)
        qs = self._get_queryset()
        fn = '.'.join([self.__field_name__, key])
        qs.update({'$unset': {fn: True}}, multi=False)
        del self._data[key]


class MapField(ContainerField):
    """ Field for maps
    """
    container = Map

    def __init__(self, value_field):
        self.value_field = value_field

    @property
    def default(self):
        return {}

    def from_mongo(self, document, value):
        return self.container(document, self, value)

    def to_mongo(self, document, value):
        return value._data



""" Common part for working with imports, documents and so on
"""

from zope.dottedname.resolve import resolve


class EnclosedDocDescriptor:
    """ Descriptor for accessing an enclosed documens within an embedded
    (:py:class:`yadm.fields.embedded.EmbeddedDocumentField`) and a reference
    (:py:class:`yadm.fields.reference.ReferenceField`) fields.

    :param str enclosed_cls_type: Enclosed class type. Can take `embedded` or
        `reference` value. Otherwise :py:exc:`ValueError` will be raised.
    """

    _DOC_CLS = 'document_class'
    _RECURSIVE_REF_CONST = 'self'

    def __init__(self, enclosed_cls_type):
        if enclosed_cls_type in ('embedded', 'reference'):
            self.attr_name = '_{}_{}'.format(enclosed_cls_type, self._DOC_CLS)
        else:
            raise ValueError

    def __get__(self, instance, owner):
        if not instance:
            return self

        value = getattr(instance, self.attr_name, None)

        if isinstance(value, str):
            if self._RECURSIVE_REF_CONST == value:
                value = getattr(instance, self._DOC_CLS)
            else:
                value = resolve(value)
            self.__set__(instance, value)

        return value

    def __set__(self, instance, value):
        setattr(instance, self.attr_name, value)

    def __delete__(self, instance):
        delattr(instance, self.attr_name)



from yadm import fields
from yadm.documents import Document, EmbeddedDocument

from .test_database import BaseDatabaseTest


class ListEmbeddedFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class EDoc(EmbeddedDocument):
            i = fields.IntegerField()
            s = fields.StringField()

        class Doc(Document):
            __collection__ = 'docs'
            li = fields.ListField(fields.EmbeddedDocumentField(EDoc))

        self.EDoc = EDoc
        self.Doc = Doc

    def test_save(self):
        td = self.Doc()

        ted = self.EDoc()
        ted.i = 13
        td.li.append(ted)

        ted = self.EDoc()
        ted.i = 42
        td.li.append(ted)

        self.db.insert(td)

        data = self.db.db.docs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 2)
        self.assertIn('i', data['li'][0])
        self.assertEqual(data['li'][0]['i'], 13)
        self.assertEqual(data['li'][1]['i'], 42)

    def test_load(self):
        self.db.db.docs.insert({'li': [{'i': 13}, {'i': 42}]})

        doc = self.db.get_queryset(self.Doc).find_one()

        self.assertTrue(hasattr(doc, 'li'))
        self.assertEqual(len(doc.li), 2)

        for item in doc.li:
            self.assertIsInstance(item, self.EDoc)

        self.assertEqual(doc.li[0].i, 13)
        self.assertEqual(doc.li[1].i, 42)

    def test_push(self):
        td = self.Doc()
        self.db.insert(td)

        ted = self.EDoc()
        ted.i = 13
        td.li.push(ted)

        ted = self.EDoc()
        ted.i = 42
        td.li.push(ted)

        data = self.db.db.docs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 2)
        self.assertIn('i', data['li'][0])
        self.assertEqual(data['li'][0]['i'], 13)
        self.assertEqual(data['li'][1]['i'], 42)

    def test_replace(self):
        td = self.Doc()
        td.li.append(self.EDoc(i=13, s='13'))
        td.li.append(self.EDoc(i=42, s='42'))
        self.db.insert(td)

        td.li.replace({'i': 13}, self.EDoc(i=26))

        data = self.db.db.docs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 2)
        self.assertIn('i', data['li'][0])
        self.assertEqual(data['li'][0]['i'], 26)
        self.assertNotIn('s', data['li'][0])
        self.assertEqual(data['li'][1]['i'], 42)
        self.assertEqual(data['li'][1]['s'], '42')

    def test_update(self):
        td = self.Doc()
        td.li.append(self.EDoc(i=13, s='13'))
        td.li.append(self.EDoc(i=42, s='42'))
        self.db.insert(td)

        td.li.update({'i': 13}, {'i': 26})

        data = self.db.db.docs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 2)
        self.assertIn('i', data['li'][0])
        self.assertEqual(data['li'][0]['i'], 26)
        self.assertEqual(data['li'][0]['s'], '13')
        self.assertEqual(data['li'][1]['i'], 42)
        self.assertEqual(data['li'][1]['s'], '42')


class DeeperListEmbeddedFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class EEDoc(EmbeddedDocument):
            i = fields.IntegerField()
            s = fields.StringField()

        class EDoc(EmbeddedDocument):
            lie = fields.ListField(fields.EmbeddedDocumentField(EEDoc))

        class Doc(Document):
            __collection__ = 'docs'
            li = fields.ListField(fields.EmbeddedDocumentField(EDoc))

        self.EEDoc = EEDoc
        self.EDoc = EDoc
        self.Doc = Doc

    def test_save(self):
        td = self.Doc()
        td.li.append(self.EDoc())

        teed = self.EEDoc()
        teed.i = 13
        td.li[0].lie.append(teed)

        self.db.save(td)

        data = self.db.db.docs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 1)
        self.assertIn('lie', data['li'][0])
        self.assertEqual(len(data['li'][0]['lie']), 1)
        self.assertIn('i', data['li'][0]['lie'][0])
        self.assertEqual(data['li'][0]['lie'][0]['i'], 13)

    def test_load(self):
        self.db.db.docs.insert({'li': [{'lie': [{'i': 13}]}]})

        doc = self.db.get_queryset(self.Doc).find_one()

        self.assertTrue(hasattr(doc, 'li'))
        self.assertEqual(len(doc.li), 1)
        self.assertTrue(hasattr(doc.li[0], 'lie'))
        self.assertEqual(len(doc.li[0].lie), 1)
        self.assertTrue(hasattr(doc.li[0].lie[0], 'i'))
        self.assertEqual(doc.li[0].lie[0].i, 13)

    def test_push(self):
        td = self.Doc()
        self.db.insert(td)

        td.li.push(self.EDoc())

        teed = self.EEDoc()
        teed.i = 13
        td.li[0].lie.push(teed)

        data = self.db.db.docs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 1)
        self.assertIn('lie', data['li'][0])
        self.assertEqual(len(data['li'][0]['lie']), 1)
        self.assertIn('i', data['li'][0]['lie'][0])
        self.assertEqual(data['li'][0]['lie'][0]['i'], 13)

    def test_pull(self):
        _id = self.db.db.docs.insert({'li': [{'lie': [{'i': 13}, {'i': 42}]}]})
        doc = self.db.get_queryset(self.Doc).find_one({'_id': _id})

        doc.li[0].lie.pull({'i': 42})

        data = self.db.db.docs.find_one({'_id': _id})

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 1)
        self.assertIn('lie', data['li'][0])
        self.assertEqual(len(data['li'][0]['lie']), 1)
        self.assertIn('i', data['li'][0]['lie'][0])
        self.assertEqual(data['li'][0]['lie'][0]['i'], 13)

    def test_replace(self):
        td = self.Doc()
        td.li.append(self.EDoc())
        td.li[0].lie.append(self.EEDoc(i=13, s='13'))
        td.li[0].lie.append(self.EEDoc(i=42, s='42'))
        self.db.insert(td)

        td.li[0].lie.replace({'i': 13}, self.EEDoc(i=26))

        data = self.db.db.docs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 1)
        self.assertIn('lie', data['li'][0])
        self.assertEqual(len(data['li'][0]['lie']), 2)
        self.assertIn('i', data['li'][0]['lie'][0])
        self.assertEqual(data['li'][0]['lie'][0]['i'], 26)
        self.assertNotIn('s', data['li'][0]['lie'][0])
        self.assertEqual(data['li'][0]['lie'][1]['i'], 42)
        self.assertEqual(data['li'][0]['lie'][1]['s'], '42')

    def test_update(self):
        td = self.Doc()
        td.li.append(self.EDoc())
        td.li[0].lie.append(self.EEDoc(i=13, s='13'))
        td.li[0].lie.append(self.EEDoc(i=42, s='42'))
        self.db.insert(td)

        td.li[0].lie.update({'i': 13}, {'i': 26})

        data = self.db.db.docs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 1)
        self.assertIn('lie', data['li'][0])
        self.assertEqual(len(data['li'][0]['lie']), 2)
        self.assertIn('i', data['li'][0]['lie'][0])
        self.assertEqual(data['li'][0]['lie'][0]['i'], 26)
        self.assertEqual(data['li'][0]['lie'][0]['s'], '13')
        self.assertEqual(data['li'][0]['lie'][1]['i'], 42)
        self.assertEqual(data['li'][0]['lie'][1]['s'], '42')



"""
Work with embedded documents.

.. code-block:: python

    class EDoc(EmbeddedDocument):
        i = fields.IntegerField()

    class Doc(Document):
        __collection__ = 'docs'
        edoc = EmbeddedDocumentField(EDoc)

    doc = Doc()
    doc.edoc = EDoc()
    doc.edoc.i = 13
    db.insert(doc)
"""

from yadm.common import EnclosedDocDescriptor
from yadm.fields.base import Field
from yadm.serialize import to_mongo, from_mongo


class EmbeddedDocumentField(Field):
    """ Field for embedded objects

    :param EmbeddedDocument embedded_document_class: class for embedded document
    """

    embedded_document_class = EnclosedDocDescriptor('embedded')

    def __init__(self, embedded_document_class):
        self.embedded_document_class = embedded_document_class

    def prepare_value(self, document, value):
        if value is None:
            return None

        elif isinstance(value, dict):
            value = self.embedded_document_class(
                __parent__=document,
                __name__=self.name,
                **value)

        elif not isinstance(value, self.embedded_document_class):
            raise TypeError('Only {!r}, dict or None is alowed, but {!r} given'
                            ''.format(self.embedded_document_class, type(value)))

        return value

    def to_mongo(self, document, value):
        if value is None:
            return None
        else:
            return to_mongo(value)

    def from_mongo(self, document, value):
        if value is None:
            return None

        elif not isinstance(value, self.embedded_document_class):
            value = from_mongo(self.embedded_document_class, value, clear_fields_changed=True)

        value.__parent__ = document
        value.__name__ = self.name
        return value

    def copy(self):
        """ Return copy of field
        """
        return self.__class__(self.embedded_document_class)



import datetime

import dateutil.parser

from yadm.fields.base import Field, DefaultMixin


class DatetimeField(DefaultMixin, Field):
    """ Field for time stamp

    :param bool auto_now: datetime.now as default
        (default: False)
    """
    def __init__(self, auto_now=False, **kwargs):
        self.auto_now = auto_now
        super().__init__(**kwargs)

    @staticmethod
    def _fix_timezone(value):
        if value.tzinfo is None:
            return value.replace(tzinfo=datetime.timezone.utc)
        else:
            return value

    @property
    def default(self):
        if self.auto_now:
            return datetime.datetime.now(datetime.timezone.utc)
        else:
            return super().default

    @classmethod
    def prepare_value(cls, document, value):
        if value is None:
            return None

        if type(value) is datetime.datetime:
            return cls._fix_timezone(value)

        elif type(value) is datetime.date:
            return datetime.datetime(
                *value.timetuple()[:3],
                tz=datetime.timezone.utc
            )

        elif isinstance(value, str):
            value = dateutil.parser.parse(value)
            return cls._fix_timezone(value)

        else:
            raise TypeError('First value must be datetime,'
                            ' date or string, but {!r}'.format(type(value)))

    @classmethod
    def from_mongo(cls, document, value):
        if value is not None:
            return cls._fix_timezone(value)

    @classmethod
    def to_mongo(cls, document, value):
        if value is not None:
            return cls._fix_timezone(value)



from setuptools import setup

# python setup.py sdist --formats=bztar

version = '0.8.3'
description = 'Yet Another Document Mapper (ODM) for MongoDB'
long_description = open('README.rst', 'rb').read().decode('utf8')


setup(
    name='yadm',
    version=version,
    description=description,
    long_description=long_description,
    author='Zelenyak "ZZZ" Aleksander',
    author_email='zzz.sochi@gmail.com',
    url='https://github.com/zzzsochi/yadm',
    license='BSD',
    platforms='any',
    install_requires=[
        'pymongo',
        'zope.dottedname',
        'python-dateutil',
    ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Database',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],

    packages=['yadm', 'yadm.fields'],
)



import random
from unittest import TestCase

from decimal import Decimal, Context, getcontext

from yadm import fields
from yadm.documents import Document

from .test_database import BaseDatabaseTest


class DecimalFieldUnitTest(TestCase):
    def test_to_mongo(self):
        field = fields.DecimalField()

        data = field.to_mongo(None, Decimal('123.45'))
        self.assertEqual(set(data.keys()), {'i', 'e'})
        self.assertEqual(data['i'], 12345)
        self.assertEqual(data['e'], -2)

    def test_from_mongo(self):
        field = fields.DecimalField()

        dec = field.from_mongo(None, {'i': 12345, 'e': -2})
        self.assertEqual(dec, Decimal('123.45'))

    def test_integer_from_digits(self):
        for _ in range(100):
            digits = [random.randint(0, 9) for _ in range(random.randint(5, 20))]
            integer = fields.DecimalField._integer_from_digits(digits)

        self.assertEqual(integer, int(''.join(str(i) for i in digits)))

    def test_func_int(self):
        field = fields.DecimalField()
        dec = field.prepare_value(None, 13)
        self.assertIsInstance(dec, Decimal)
        self.assertEqual(dec, Decimal(13))

    def test_func_str(self):
        field = fields.DecimalField()
        dec = field.prepare_value(None, '3.14')
        self.assertIsInstance(dec, Decimal)
        self.assertEqual(dec, Decimal('3.14'))

    def test_func_error(self):
        field = fields.DecimalField()
        self.assertRaises(TypeError, field.prepare_value, type)

    def test_context_default(self):
        field = fields.DecimalField()
        self.assertIsInstance(field.context, Context)
        self.assertEqual(field.context, getcontext())

    def test_context_init(self):
        context = Context(prec=5)
        field = fields.DecimalField(context=context)
        self.assertIsInstance(field.context, Context)
        self.assertEqual(field.context, context)


class DecimalFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdocs'
            dec = fields.DecimalField()

        self.TestDoc = TestDoc

    def test_save(self):
        td = self.TestDoc()
        td.dec = Decimal('3.14')
        self.db.save(td)

        data = self.db.db.testdocs.find_one()

        self.assertIn('dec', data)
        self.assertEqual(set(data['dec']), {'i', 'e'})
        self.assertEqual(data['dec']['i'], 314)
        self.assertEqual(data['dec']['e'], -2)

    def test_load(self):
        self.db.db.testdocs.insert({'dec': {'i': 314, 'e': -2}})

        doc = self.db.get_queryset(self.TestDoc).find_one()

        self.assertTrue(hasattr(doc, 'dec'))
        self.assertIsInstance(doc.dec, Decimal)
        self.assertEqual(doc.dec, Decimal('3.14'))
